import { ICollection, ICompare } from '@algorithm.ts/types';

interface IQueue<T> extends ICollection<T> {
    /**
     * Initialize queue with initial elements.
     * @param elements
     * @param start
     * @param end
     */
    init(elements?: ReadonlyArray<T>, start?: number, end?: number): void;
    /**
     * Add an element to the tail of the queue.
     * @param element
     */
    enqueue(element: T): void;
    /**
     * Add multiple elements to the tail of the queue.
     * @param elements
     * @param start
     * @param end
     */
    enqueues(elements: ReadonlyArray<T>, start?: number, end?: number): void;
    /**
     * Popup the front element from the queue.
     * @param newElement if specified, the element will be enqueued, this parameter is just for better performance.
     */
    dequeue(newElement?: T): T | undefined;
    /**
     * Remove existed elements which is not accepted by the filter, then enqueues the additional elements.
     * @param filter
     * @param elements
     * @param start         start index of `newElements`
     * @param end           end index of `newElements`
     */
    splice(filter: (element: T) => boolean, elements?: ReadonlyArray<T>, start?: number, end?: number): void;
    /**
     * Get the front element of the queue.
     */
    front(): T | undefined;
}
interface IDeque<T> extends IQueue<T> {
    /**
     * Popup the last element of the queue.
     */
    pop(): T | undefined;
    /**
     * Get the last element of the queue.
     */
    back(): T | undefined;
    /**
     * Add an element to the head of the queue.
     */
    unshift(element: T): void;
}
/**
 * Circular queue.
 *
 * Circular queue is a queue structure, the main purpose of its design is to
 * reuse space as much as possible on the basis of ordinary queues. Circular
 * queues usually need to specify the maximum volume C of the collector. If the
 * number of elements in the queue exceeds C, only the most recent C elements
 * are kept in the queue. Other operations are the same as ordinary queues.
 */
interface ICircularQueue<T> extends IQueue<T>, IDeque<T> {
    /**
     * Resize the max-size of queue with the given size.
     * @param MAX_SIZE
     */
    resize(MAX_SIZE: number): void;
    /**
     * Rearrange elements, that is, put the first element in place 0-index.
     */
    rearrange(): void;
}
/**
 * Priority queue.
 *
 * Priority queue can insert elements within the complexity of log(N) or remove
 * the largest element in the queue.
 */
type IPriorityQueue<T> = IQueue<T>;

interface ICircularQueueProps {
    /**
     * Initial capacity of the circular queue.
     */
    capacity?: number;
    /**
     * Automatically extends the queue capacity when the queue is full.
     * @default true
     */
    autoResize?: boolean;
    /**
     * @default 1.5
     */
    autoResizeExpansionRatio?: number;
}
declare class CircularQueue<T = unknown> implements ICircularQueue<T> {
    protected readonly _elements: T[];
    protected readonly _autoResize: (nextExpectedSize: number) => void;
    protected _capacity: number;
    protected _size: number;
    protected _start: number;
    protected _end: number;
    constructor(props?: ICircularQueueProps);
    [Symbol.iterator](): Iterator<T>;
    init(elements?: ReadonlyArray<T>, start?: number, end?: number): void;
    resize(MAX_SIZE: number): void;
    enqueue(element: T): void;
    unshift(element: T): void;
    enqueues(elements: ReadonlyArray<T>, start?: number, end?: number): void;
    dequeue(element?: T): T | undefined;
    pop(): T | undefined;
    splice(filter: (element: T) => boolean, elements?: ReadonlyArray<T> | undefined, start?: number | undefined, end?: number | undefined): void;
    front(): T | undefined;
    back(): T | undefined;
    rearrange(): void;
    destroy(): void;
    clear(): void;
    get size(): number;
    protected _forwardIndex(index: number): number;
    protected _backwardIndex(index: number): number;
}

interface IPriorityQueueProps<T> {
    /**
     * If the compare(x, y) < 0, then x has a higher precedence than y.
     */
    compare: ICompare<T>;
}
/**
 * MinHeap
 */
declare class PriorityQueue<T> implements IPriorityQueue<T> {
    protected readonly _compare: ICompare<T>;
    protected readonly _elements: T[];
    protected _size: number;
    constructor(props: IPriorityQueueProps<T>);
    [Symbol.iterator](): Iterator<T>;
    init(elements?: ReadonlyArray<T>, start?: number, end?: number): void;
    enqueue(element: T): void;
    enqueues(elements: ReadonlyArray<T>, start?: number, end?: number): void;
    dequeue(element?: T): T | undefined;
    splice(filter: (element: T) => boolean, elements?: ReadonlyArray<T>, start?: number, end?: number): void;
    front(): T | undefined;
    destroy(): void;
    clear(): void;
    get size(): number;
    protected _down(index: number): void;
    protected _up(index: number): void;
    protected _fastBuild(): void;
}

export { CircularQueue, ICircularQueue, ICircularQueueProps, IDeque, IPriorityQueue, IPriorityQueueProps, IQueue, PriorityQueue };
